/* Copyright (C) 2002, by Rupert Paget (rpaget@vision.ee.ethz.ch),
 * all rights reserved.
 *
 * This software is copyrighted as noted below. It may be freely copied,
 * modified, and redistributed, provided that the copyright notice is
 * preserved on all copies.
 *
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is". Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a commercial program or other
 * software product without the expressed permission of the copyright owner.
 *
 * If you modify this software, you should include a notice giving the
 * name of the person performing the modification, the date of modification,
 * and the reason for such modification.
 */

#include "Option.h"

/**************************************************************************/

Option::Option() {
  inputFileVn = 0;
  inputFileMaskVn = 0;
  maskIn = false;
  maskOut = false;
  levels = LEVELS;
  neighOrder = NEIGH_ORDER;
  square = SQUARE;
  treeMax = TREE_MAX;
  cyclic = CYCLIC;
  iterations = ITERATIONS;
  window = WINDOW;
  kNearest = KNEAREST;
  rows = 0;
  cols = 0;
}

/**************************************************************************/

Option::Option(int theArgc, const char **theArgv) {
  Option();
  parseArgs(theArgc,theArgv);
}

/**************************************************************************/

Option::~Option() {
}

/**************************************************************************/

void Option::parseArgs(int theArgc, const char **theArgv)
{
  const char *progName = *theArgv; //store program name
  theArgc--; theArgv++;
  
  /* exit if no parameters */
  if (theArgc < 1) help(progName);
  
  /* process options */
  while (theArgc) {
    if (**theArgv == '-') {
      switch((*theArgv)[1]){
      case 'l':
	theArgc--;
	levels = atoi(*++theArgv);
	if (levels < 0) {
	  cerr << "Error: number of levels < 0" << endl;
	  exit(1);
	}
	if (levels > LEVELS) levels = LEVELS;
	break;
      case 'n':
	theArgc--;
	neighOrder = atoi(*++theArgv);
	if (neighOrder < 0) {
	  cerr << "Error: neighOrder < 0" << endl;
	  exit(1);
	}
	break;
      case 's':
	square = true;
	break;
      case 't':
	theArgc--;
	treeMax = atoi(*++theArgv);
	if (treeMax < 0) {
	  cerr << "Error: treeMax < 0" << endl;
	  exit(1);
	}
	break;
      case 'i':
	theArgc--;
	iterations = atoi(*++theArgv);
	if (iterations <= 0) {
	  cerr << "Error: number of iterations <= 0" << endl;
	  exit(1);
	}
	break;
      case 'w':
	theArgc--;
	window = atoi(*++theArgv);
	if (window < 0) {
	  cerr << "Error: window width < 0" << endl;
	  exit(1);
	}
	break;	
      case 'k':
	theArgc--;
	kNearest = atoi(*++theArgv);
	if (kNearest < 0) {
	  cerr << "Error: kNearest < 0" << endl;
	  exit(1);
	}
	break;	
      case 'c':
	cyclic = true;
	break;
      case 'x':
	theArgc--;
	cols = atoi(*++theArgv);
	if (cols <= 0) {
	  cerr << "Error: number of cols <= 0" << endl;
	  exit(1);
	}
	break;
      case 'y':
	theArgc--;
	rows = atoi(*++theArgv);
	if (rows <= 0) {
	  cerr << "Error: number of rows <= 0" << endl;
	  exit(1);
	}
	break;
      case 'm':
	theArgc--;
	inputFileMaskV.push_back(*++theArgv);
	break;
      case 'h':
	help(progName);
	break;
      default :
	cerr << "Invalid option " << *theArgv << endl;
	help(progName);
	break;
      }
    } else {
      inputFileV.push_back(*theArgv);
    }
    theArgc--; theArgv++;
  }   /* while theArgc */
  
  if (inputFileV.empty())
    cerr << "Youch! No input texture file specified." << endl;
    
} /* parseArgs() */

/**************************************************************************/

void Option::help(const char *progName) {
  cerr << endl << endl;
  cerr << "USAGE: " << progName << " [-l levels] [-n neighOrder] [-square]";
  cerr << endl;
  cerr << "       [-t treeMax] [-cyclic] [-x cols] [-y rows] [-m maskfile]";
  cerr << endl;
  cerr << "       [-help] <texturefile>" << endl << endl;

  cerr << "DEFAULT: levels = " << levels << ", neighOrder = " << neighOrder;
  cerr << ", square = " << square << ", treeMax = " << treeMax << endl;
  cerr << "         cyclic = " << cyclic << ", cols = " << cols;
  cerr << ", rows = " << rows << endl << endl;
  exit (1);
} /* help() */

/**************************************************************************/

void Option::setRows(int rowsOut) {
  rows = rowsOut;
}

/**************************************************************************/

void Option::setCols(int colsOut) {
  cols = colsOut;
}

/**************************************************************************/

void Option::setTreeMax(int treeMaxIn) {
  treeMax = treeMaxIn;
}

/**************************************************************************/

bool Option::InputFile() {
  if (inputFileVn >= inputFileV.size()) return false;
  inputFile = inputFileV[inputFileVn++];
  if (inputFileMaskVn >= inputFileMaskV.size()) maskIn = false;
  else {
    inputFileMask = inputFileMaskV[inputFileMaskVn++];
    maskIn = true;
  }
  return true;
}

/**************************************************************************/

bool Option::OutputFile() {
  FILE *fp;
  char addName[500];
  addName[0] = '\0';
  if (levels != LEVELS) sprintf(addName,"_l%d",levels);
  sprintf(addName,"%s_n%d",addName,neighOrder);
  if (square) sprintf(addName,"%s_s",addName);
  sprintf(addName,"%s_t%d",addName,treeMax);
  if (iterations != ITERATIONS) sprintf(addName,"%s_i%d",addName,iterations);
  if (window != WINDOW) sprintf(addName,"%s_w%d",addName,window);
  if (kNearest != KNEAREST) sprintf(addName,"%s_k%d",addName,kNearest);
  if (cyclic) sprintf(addName,"%s_c",addName);
  sprintf(addName,"%s_%dx%d",addName,cols,rows);
  outputFile = gstrnew(inputFile.c_str(),addName);

  /* Check if output file already exists */
  if ((fp = fopen(outputFile.c_str(),"r")) == NULL) {
    fp = fopen(outputFile.c_str(),"w");
    fclose(fp);
    return false;
  }
  fclose(fp);

  /* Output file already exists */
  cerr << endl;
  cerr << "Default output file name \"" << outputFile.c_str();
  cerr << "\" already exists." << endl;
  cerr << "This program will NOT overwrite an existing file." << endl;
  cerr << "Move or remove this file if you wish to obtain a new image.";
  cerr << endl << endl;
  return true;
}

/**************************************************************************/

