/* Copyright (C) 2002, by Rupert Paget (rpaget@vision.ee.ethz.ch),
 * all rights reserved.
 *
 * This software is copyrighted as noted below. It may be freely copied,
 * modified, and redistributed, provided that the copyright notice is
 * preserved on all copies.
 *
 * There is no warranty or other guarantee of fitness for this software,
 * it is provided solely "as is". Bug reports or fixes may be sent
 * to the author, who may or may not act on them as he desires.
 *
 * You may not include this software in a commercial program or other
 * software product without the expressed permission of the copyright owner.
 *
 * If you modify this software, you should include a notice giving the
 * name of the person performing the modification, the date of modification,
 * and the reason for such modification.
 */

#include <iostream>
#include <sys/types.h>
#include <time.h>
#include <list>
#include <vector>
#include "ColourImage.h"
#include "GreyImage.h"
#include "Option.h"
#include "Pyramid.h"
#include "Neighbourhood.h"
extern "C" {
#include <nr.h>
}

using namespace std;

#define pMax 255
// #define dMax 195075    // 3 x pMax^2, for Euclidean distance
#define dMax 765       // 3 x pMax, for Manhattan distance

BasicImage::byte pUpdateFn(int cO, int rO, GreyImage &imgProb, int increment);
list<int> *SimilarPixels(Pyramid *pyramid, int t, Neighbourhood &n,
			 Option &opt);
void IndexArrayBasic(list<int> &index, int cOn, int rOn, int dx, int dy,
		     int dAx, int dAy, int dt, int t, Pyramid *pyramid);
void IndexArray(list<int> &index, int cOn, int rOn, int dx, int dy,
		int dAx, int dAy, int dt, int t, Pyramid *pyramid);
void IndexArray2(list<int> &index, int cI, int rI, int t, Pyramid *pyramid);
void IndexArray3(list<int> &index, int cI, int rI, int t, Pyramid *pyramid);
bool getImgPos(int &c, int &r, list<int> &index, int cols, int rows,
	       GreyImage &imgMask, bool mask);
bool getImgPos2(int &c, int &r, int startIndex, list<int> &index,
		int cols, int rows, GreyImage &imgMask, bool mask);


int main(int argc, const char *argv[]) {
  
  /* Warm up random number generator */
  long iseed;
  time(&iseed);
  iseed = -iseed;
  for (int i = 0; i < 1000; ++i) ran3(&iseed);

  /* Read in arguments */
  Option opt;
  opt.parseArgs(argc, argv);
  while(opt.InputFile()) {
    
    /* Set neighbourhood */
    Neighbourhood n(opt.getNeighOrder(),opt.getSquare(),true);
    
    /* Build pyramid */
    Pyramid *pyramid = new Pyramid [opt.getTreeMax()+1];
    pyramid[0].load(opt);
    int t = 1;
    for (; t <= opt.getTreeMax(); ++t){
      if (!pyramid[t].reduce(pyramid[t-1])) break;
    }
    if (t <= opt.getTreeMax()) opt.setTreeMax(t-1);

    /* Set output file */
    if (opt.OutputFile()) continue;
    
    /* Malloc variables */
    int nLength = n.getlength();
    bool *pOkN = new bool [nLength];
    float *pForN = new float [nLength];
    BasicImage::byte **imgOutN = new BasicImage::byte* [nLength];
    imgOutN[0] = new BasicImage::byte [nLength*3];
    for (int i = 1; i < nLength; ++i)  imgOutN[i] = imgOutN[i-1]+3;

    /* Start iteration from scale treeMax down */
    vector<int> rInMin;
    vector<int> cInMin;
    for (t = opt.getTreeMax(); t >= 0; --t){
      Pyramid &pyr = pyramid[t];

      /* Input image indexing */
      pyr.imgInIndexl = SimilarPixels(pyramid,t,n,opt);

      /* Determine total number of points to iterate */
      float probTotal = pMax*pyr.rowsOut*pyr.colsOut;
      
      /* Random loop through output image */
      int rO, cO;
      int pIncrement = pMax/opt.getIterations()+1;
      while (probTotal > 0) {
	float p = ran3(&iseed)*probTotal;
	for (rO = 0; (rO < pyr.rowsOut) && (p > 0.0); ++rO)
	for (cO = 0; (cO < pyr.colsOut) && (p > 0.0); ++cO)
	  p -= pMax-pyr.imgProb.p(cO,rO);
	if (rO > 0) --rO;
	if (cO > 0) --cO;
	
	/* Update probabilities */
	BasicImage::byte pUpdate = pUpdateFn(cO,rO,pyr.imgProb,pIncrement);
	pyr.imgProb.setp(cO,rO,pUpdate+pyr.imgProb.p(cO,rO));
	probTotal -= pUpdate;
	
	/* Setup constant Neighbourhood vectors */
	int rOt = rO/2;
	int cOt = cO/2;
	int dAy = rO%2;
	int dAx = cO%2;
	int nPos = -1;
	int dx, dy, dt;
	list<int> index;
	list<int> indexBasic;
	while (n.neigh(dx,dy,dt)) {
	  int rOn, cOn;
	  if (dt) {
	    if (t == opt.getTreeMax()) continue;
	    rOn = rOt+dy;
	    cOn = cOt+dx;
	  } else {
	    rOn = rO+dy;
	    cOn = cO+dx;
	  }
	  ++nPos;
	  Pyramid &pyr = pyramid[t+dt];
	  
	  /* If output images are toroidal */
	  if (opt.getCyclic()) {
	    rOn = (pyr.rowsOut + (rOn % pyr.rowsOut)) % pyr.rowsOut;
	    cOn = (pyr.colsOut + (cOn % pyr.colsOut)) % pyr.colsOut;
	  }
	  /* Else check suitability */
	  else if ((rOn < 0) || (rOn >= pyr.rowsOut) ||
		   (cOn < 0) || (cOn >= pyr.colsOut)) {
	    pOkN[nPos] = false;
	    continue;
	  }
	  
	  if (dt) {
	    /* Probability factors */
	    pOkN[nPos] = true;
	    pForN[nPos] = 1.0;

	  } else {	    
	    /* Get neighbour pixel probability */
	    BasicImage::byte pn = pyr.imgProb.p(cOn,rOn);
	    if (pn == 0) {
	      pOkN[nPos] = false;
	      continue;
	    }
	    
	    /* Probability factors */
	    pOkN[nPos] = true;
	    pForN[nPos] = float(pn)/pMax;
	  }

	  /* Store neighbourhood variables */
	  for (int band = 0; band < 3; ++band)
	    imgOutN[nPos][band] = pyr.imgOut.getBand(band)->p(cOn,rOn);

	  /* Update input image indexes */
	  IndexArrayBasic(indexBasic,cOn,rOn,dx,dy,dAx,dAy,dt,t,pyramid);
	  IndexArray(index,cOn,rOn,dx,dy,dAx,dAy,dt,t,pyramid);
	}

	/* If indexBasic has only one pixel, use it. */
	/* Otherwise search for pixel in index. */
	if (indexBasic.size() == 1) {
	  list<int>::iterator i = indexBasic.begin();
	  int cI = (*i)%pyr.colsIn;
	  int rI = (*i)/pyr.colsIn;
	  
	  /* Set new output pixel */
	  for (int band = 0; band < 3; ++band)
	    pyr.imgOut.getBand(band)->
	      setp(cO,rO,pyr.imgIn.getBand(band)->p(cI,rI));
	
	  /* Set new output index */
	  pyr.imgOutIndex[cO][rO] = pyr.imgInIndexl[*i].front();
	  continue;
	}
	
	/* Search input texture for pixel with minumum */
	/* neighbourhood distance */
	int rI, cI;
	float NeighMin = FLT_MAX;
	while (getImgPos(cI,rI,index,pyr.colsIn,pyr.rowsIn,
			 pyr.imgInMask,opt.getMaskIn())) {
	  int rIt = rI/2;
	  int cIt = cI/2;
	  nPos = -1;
	  float Neigh = 0.0;
	  while (n.neigh(dx,dy,dt,Neigh,NeighMin)) {
	    int rIn, cIn;
	    if (dt) {
	      if (t == opt.getTreeMax()) continue;
	      rIn = rIt+dy;
	      cIn = cIt+dx;
	    } else {
	      rIn = rI+dy;
	      cIn = cI+dx;
	    }
	    ++nPos;
	    Pyramid &pyr = pyramid[t+dt];

	    /* Probability factor */
	    float pFor = pForN[nPos];
	    // pFor *= pFor; // For Euclidean distance
	    
	    /* Check suitability */
	    if (!pOkN[nPos]) continue;
	    if ((rIn < 0) || (rIn >= pyr.rowsIn) ||
		(cIn < 0) || (cIn >= pyr.colsIn) ||
		(opt.getMaskIn() && !pyr.imgInMask.p(cIn,rIn))) {
	      Neigh += dMax*pFor;
	      continue;
	    }
	      
	    /* Calculate Manhatten neighbourhood distance */
	    /* Calculate Euclidean neighbourhood distance */
	    int d;
	    for (int band = 0; band < 3; ++band) {
	      d = imgOutN[nPos][band]-pyr.imgIn.getBand(band)->p(cIn,rIn);
	      if (d < 0) Neigh -= d*pFor;
	      else Neigh += d*pFor;
	      // Neigh += d*d*pFor;
	    }
	  }
	  
	  /* Store neighbour pixel distance */
	  if (Neigh < NeighMin) {
	    NeighMin = Neigh;
	    rInMin.clear();
	    cInMin.clear();
	    rInMin.push_back(rI);
	    cInMin.push_back(cI);
	  } else if (Neigh == NeighMin) {
	    rInMin.push_back(rI);
	    cInMin.push_back(cI);
	  }
	}
	
	/* Sample new output pixel */
	int vLength = rInMin.size();
	if (vLength == 1) {
	  rI = rInMin[0];
	  cI = cInMin[0];
	} else {
	  int i = int(ran3(&iseed)*vLength);
	  if (i == vLength) --i;
	  rI = rInMin[i];
	  cI = cInMin[i];
	}
	
	/* Set new output pixel */
	for (int band = 0; band < 3; ++band)
	  pyr.imgOut.getBand(band)->
	    setp(cO,rO,pyr.imgIn.getBand(band)->p(cI,rI));
	
	/* Set new output index */
	pyr.imgOutIndex[cO][rO] = pyr.imgInIndexl[cI+rI*pyr.colsIn].front();
      }
    }
  
    /* Delete variables */
    delete [] pOkN;
    delete [] pForN;
    delete [] imgOutN[0];
    delete [] imgOutN;
    
    /* Ouput synthesized texture */
    pyramid[0].imgOut.save(opt.getOutputFile());
    
  }
  return 0;
}

/**************************************************************************/

BasicImage::byte pUpdateFn(int cO, int rO, GreyImage &imgProb, int inc) {
  // pUpdateFn MUST return a value x such that 0 < x <= pMax-imgProb.p(cO,rO)
  // and whereby x will equal pMax-imgProb.p(cO,rO) at some point in time.
  BasicImage::byte p = imgProb.p(cO,rO);
  if (inc < pMax-p) return inc;
  return pMax-p;
}

/**************************************************************************/

list<int> *SimilarPixels(Pyramid *pyramid, int t, Neighbourhood &n,
			 Option &opt) {
  // Return an image of lists, where each list is list of indexes to
  // similar pixels for the pixel of concern. The lists are no longer
  // than k. Similarity is determine by the L1 (Manhatten) distance
  // between neighbourhoods. Note even if k = 0, there will always be
  // at least one pixel in the list, as each pixel is inserted into its
  // own list.
  //
  // As the index is a single int,
  // row = index/cols;
  // col = index%cols;
  //
  Pyramid &pyr = pyramid[t];

  /* Setup output index image */
  list<int> *imgIndex = new list<int> [pyr.colsIn*pyr.rowsIn];
  list<float> *imgIndexN = new list<float> [pyr.colsIn*pyr.rowsIn];
  
  // Instead of collecting a set of indexes for each pixel,
  // the first index will indicate which set of indexes the pixel
  // should refer to. This will stop the repeatitions of sets.
  // Also speed can be improved, as once a pixel has been redirected
  // to a set, it does not need to be used in further similarity comparisons.
  // A pixel can be only redirected to a set, if the L1 (Manhatten) 
  // distance between neighbourhoods = 0.
  bool *pOkN = new bool [n.getlength()];
  BasicImage::byte **imgN = new BasicImage::byte* [n.getlength()];
  imgN[0] = new BasicImage::byte [3*n.getlength()];
  for (int i = 1; i < n.getlength(); ++i) imgN[i] = imgN[i-1]+3;

  /* Initialize each pixel index */
  int k = opt.getKNearest();
  if (k < 0) k=1;
  else k++;
  for (int r = 0; r < pyr.rowsIn; ++r) {
  for (int c = 0; c < pyr.colsIn; ++c) {
    if (!opt.getMaskIn() || pyr.imgInMask.p(c,r)) {
      int index=c+r*pyr.colsIn;
      imgIndex[index].push_back(index);
      imgIndexN[index].push_back(0.0);
    }
  }
  }

  /* For each pixel */
  for (int r = 0; r < pyr.rowsIn; ++r) {
  for (int c = 0; c < pyr.colsIn; ++c) {
    if (!opt.getMaskIn() || pyr.imgInMask.p(c,r)) {
      int index=c+r*pyr.colsIn;
      
      /* Check that index has not allready been redirected */
      if (imgIndex[index].front() != index) continue;
      
      /* Setup constant Neighbourhood vectors */
      int rt = r/2;
      int ct = c/2;
      int nPos = -1;
      int dx, dy, dt;
      while (n.neigh(dx,dy,dt)) {
	int rn, cn;
	if (dt) {
	  if (t == opt.getTreeMax()) continue;
	  rn = rt+dy;
	  cn = ct+dx;
	} else {
	  rn = r+dy;
	  cn = c+dx;
	}
	++nPos;
	Pyramid &pyr = pyramid[t+dt];

	/* Check suitability */
	if ((rn < 0) || (rn >= pyr.rowsIn) || (cn < 0) || (cn >= pyr.colsIn) ||
	    (opt.getMaskIn() && !pyr.imgInMask.p(cn,rn))) {
	  pOkN[nPos] = false;
	  continue;
	}

	/* Store neighbourhood variables */
	pOkN[nPos] = true;
	for (int band = 0; band < 3; ++band)
	  imgN[nPos][band] = pyr.imgIn.getBand(band)->p(cn,rn);
      }

      /* Setup search index */
      /* note: number of direct comparisons needed = size*(size-1)/2 */
      /* whereas number of comparisons done via indexList = size*4*k */
      //list<int> indexList;
      //if ((t < opt.getTreeMax()) && (pyr.colsIn*pyr.rowsIn > 8*k))
      //	IndexArray2(indexList,c,r,t,pyramid);
      
      /* Setup search index */
      /* note: number of direct comparisons needed = size*(size-1)/2 */
      /* whereas number of comparisons done via indexList = size*4*k*k */
      list<int> indexList;
      if ((t < opt.getTreeMax()) && (pyr.colsIn*pyr.rowsIn > 8*k*k))
	IndexArray3(indexList,c,r,t,pyramid);
      
      /* Search forward */
      float NeighMin1 = FLT_MAX;
      int size1 = imgIndex[index].size();
      if (size1 >= k) NeighMin1 = imgIndexN[index].back();
      int ci, ri;
      while (getImgPos2(ci,ri,index,indexList,pyr.colsIn,pyr.rowsIn,
			pyr.imgInMask,opt.getMaskIn())) {
	int indexi = ci+ri*pyr.colsIn;
	float NeighMin2 = FLT_MAX;
	int size2 = imgIndex[indexi].size();
	if (size2 >= k) NeighMin2 = imgIndexN[indexi].back();
	float NeighMin = (NeighMin1 > NeighMin2) ? NeighMin1:NeighMin2;
	
	rt = ri/2;
	ct = ci/2;
	nPos = -1;
	float Neigh = 0.0;
	while (n.neigh(dx,dy,dt,Neigh,NeighMin)) {
	  int rn, cn;
	  if (dt) {
	    if (t == opt.getTreeMax()) continue;
	    rn = rt+dy;
	    cn = ct+dx;
	  } else {
	    rn = ri+dy;
	    cn = ci+dx;
	  }
	  ++nPos;
	  Pyramid &pyr = pyramid[t+dt];
	  
	  /* Check suitability */
	  if (!pOkN[nPos]) {
	    Neigh += dMax;
	    continue;
	  }
	  if ((rn < 0) || (rn >= pyr.rowsIn) ||
	      (cn < 0) || (cn >= pyr.colsIn) ||
	      (opt.getMaskIn() && !pyr.imgInMask.p(cn,rn))) {
	    Neigh += dMax;
	    continue;
	  }
	  
	  /* Calculate Manhatten neighbourhood distance */
	  int d;
	  for (int band = 0; band < 3; ++band) {
	    d = imgN[nPos][band]-pyr.imgIn.getBand(band)->p(cn,rn);
	    if (d < 0) Neigh -= d;
	    else Neigh += d;
	  }
	}
	
	/* Store new index */
	if (Neigh < NeighMin1) {
	  list<int>::iterator i = imgIndex[index].begin();
	  list<float>::iterator iN = imgIndexN[index].begin();
	  for (;i != imgIndex[index].end() && *iN <= Neigh; ++i, ++iN);
	  imgIndex[index].insert(i,indexi);
	  imgIndexN[index].insert(iN,Neigh);
	  size1++;
	  if (size1 > k) {
	    imgIndex[index].pop_back();
	    imgIndexN[index].pop_back();
	    NeighMin1 = imgIndexN[index].back();
	    size1--;
	  }
	}
	if (Neigh < NeighMin2) {
	  list<int>::iterator i = imgIndex[indexi].begin();
	  list<float>::iterator iN = imgIndexN[indexi].begin();
	  for (;i != imgIndex[indexi].end() && *iN <= Neigh; ++i, ++iN);
	  imgIndex[indexi].insert(i,index);
	  imgIndexN[indexi].insert(iN,Neigh);
	  size2++;
	  if (size2 > k) {
	    imgIndex[indexi].pop_back();
	    imgIndexN[indexi].pop_back();
	  }
	}
      }
    }
  }
  }

  delete [] pOkN;
  delete [] imgN[0];
  delete [] imgN;
  delete [] imgIndexN;
  return imgIndex;
}

/**************************************************************************/

void IndexArrayBasic(list<int> &index, int cOn, int rOn, int dx, int dy,
		     int dAx, int dAy, int dt, int t, Pyramid *pyramid) {
  // Update index array with new indexes, making sure that there are no
  // repeatitions and they do not point to points outside the image.
  Pyramid &pyr = pyramid[t];
  Pyramid &pyrUp = pyramid[t+dt];

  /* If imgOutIndex is neighbourhood pixel then the center pixel is */
  int r = (pyrUp.imgOutIndex[cOn][rOn]/pyrUp.colsIn)-dy;
  int c = (pyrUp.imgOutIndex[cOn][rOn]%pyrUp.colsIn)-dx;
  if (dt) {
    r = 2*r+dAy;
    c = 2*c+dAx;
  }
  if ((r < 0) || (c < 0) || (c >= pyr.colsIn) || (r >= pyr.rowsIn)) return;
  int j = c+r*pyr.colsIn;
  list<int>::const_iterator im = pyr.imgInIndexl[j].begin();
  if (im == pyr.imgInIndexl[j].end()) return;
  list<int> &imgIndex = pyr.imgInIndexl[*im];
  im = imgIndex.begin();

  /* First index */
  /* Add imgIndex if it does not already exist */
  list<int>::iterator i;
  for (i = index.begin(); i != index.end(); ++i) {
    if (*i == *im) break;
    if (*i > *im) {
      index.insert(i,*im);
      break;
    }
  }
  if (i == index.end()) index.push_back(*im);
}

/**************************************************************************/

void IndexArray(list<int> &index, int cOn, int rOn, int dx, int dy,
		int dAx, int dAy, int dt, int t, Pyramid *pyramid) {
  // Update index array with new indexes, making sure that there are no
  // repeatitions and they do not point to points outside the image.
  // The indexes are listed in canological order to allow quick correspondence.
  Pyramid &pyr = pyramid[t];
  Pyramid &pyrUp = pyramid[t+dt];

  /* If imgOutIndex is neighbourhood pixel then the center pixel is */
  int r = (pyrUp.imgOutIndex[cOn][rOn]/pyrUp.colsIn)-dy;
  int c = (pyrUp.imgOutIndex[cOn][rOn]%pyrUp.colsIn)-dx;
  if (dt) {
    r = 2*r+dAy;
    c = 2*c+dAx;
  }
  if ((r < 0) || (c < 0) || (c >= pyr.colsIn) || (r >= pyr.rowsIn)) return;
  int j = c+r*pyr.colsIn;
  list<int>::const_iterator im = pyr.imgInIndexl[j].begin();
  if (im == pyr.imgInIndexl[j].end()) return;
  list<int> &imgIndex = pyr.imgInIndexl[*im];
  im = imgIndex.begin();

  /* First index */
  /* Add imgIndex if it does not already exist */
  list<int>::iterator i;
  for (i = index.begin(); i != index.end(); ++i) {
    if (*i == *im) break;
    if (*i > *im) {
      index.insert(i,*im);
      break;
    }
  }
  if (i == index.end()) index.push_back(*im);

  /* The rest of the indexes are in canological order */
  i = index.begin();
  for (++im; im != imgIndex.end(); ++im) {
    /* Add imgIndex if it does not already exist */
    for (; i != index.end(); ++i) {
      if (*i == *im) break;
      if (*i > *im) {
	index.insert(i,*im);
	break;
      }
    }
    if (i == index.end()) index.push_back(*im);
  }
}

/**************************************************************************/

void IndexArray2(list<int> &index, int cI, int rI, int t, Pyramid *pyramid) {
  // This subroutine takes the previous list of indexes from the above level
  // to create a new list of possible indexes from which the k nearest
  // neighbours will be found. In the current implementation, the new list
  // will only be about 4*k big, however this could be simply increased to
  // 4*k*k by considering also the lists at the referenced cites. This would
  // increase the search time while finding a "better" set of k nearest
  // neighbours, but there would not be an increase in memory requirments.
  // See IndexArray3 for implementation.
  Pyramid &pyr = pyramid[t];
  Pyramid &pyrUp = pyramid[t+1];

  int pos = cI+rI*pyr.colsIn;
  int j = cI/2+rI/2*pyrUp.colsIn;
  list<int>::const_iterator im = pyrUp.imgInIndexl[j].begin();
  if (im == pyrUp.imgInIndexl[j].end()) return;
  list<int> &imgIndex = pyrUp.imgInIndexl[*im];
  
  /* For each imgIndex at designated point in higher pyramid */
  list<int>::iterator i;
  for (im = imgIndex.begin(); im != imgIndex.end(); ++im) {
    int c = 2*(*im%pyrUp.colsIn);
    int r = 2*(*im/pyrUp.colsIn);
    for (int dy = 0; dy < 2; ++dy) {
    for (int dx = 0; dx < 2; ++dx) {
      int ci = c+dx;
      int ri = r+dy;
      j = ci+ri*pyr.colsIn;
      if ((j == pos) || (ci >= pyr.colsIn) || (ri >= pyr.rowsIn) || 
	  (pyr.getMaskIn() && !pyr.imgInMask.p(ci,ri))) continue;

      /* Add imgIndex if it does not already exist */
      for (i = index.begin(); i != index.end(); ++i) {
	if (*i == j) break;
	if (*i > j) {
	  index.insert(i,j);
	  break;
	}
      }
      if (i == index.end()) index.push_back(j);
    }
    }
  }
}

/**************************************************************************/

void IndexArray3(list<int> &index, int cI, int rI, int t, Pyramid *pyramid) {
  // This subroutine takes the previous list of indexes from the above level
  // to create a new list of possible indexes from which the k nearest
  // neighbours will be found. This is an updated implementation of IndexArray2,
  // whereby the first list of k nearest neighbours is used to access each sites
  // own list of k nearest neighbours. This gives a maximum of 4*k*k sites are
  // considered. This will increase the search time while finding a "better"
  // set of k nearest neighbours. Note there is not be an increase in memory
  // requirments.
  Pyramid &pyr = pyramid[t];
  Pyramid &pyrUp = pyramid[t+1];

  int pos = cI+rI*pyr.colsIn;
  int j = cI/2+rI/2*pyrUp.colsIn;
  list<int>::const_iterator im = pyrUp.imgInIndexl[j].begin();
  if (im == pyrUp.imgInIndexl[j].end()) return;
  list<int> &imgIndex = pyrUp.imgInIndexl[*im];
  
  /* For each imgIndex at designated point in higher pyramid */
  list<int>::iterator i;
  for (im = imgIndex.begin(); im != imgIndex.end(); ++im) {
    int c = 2*(*im%pyrUp.colsIn);
    int r = 2*(*im/pyrUp.colsIn);
    for (int dy = 0; dy < 2; ++dy) {
    for (int dx = 0; dx < 2; ++dx) {
      int ci = c+dx;
      int ri = r+dy;
      j = ci+ri*pyr.colsIn;
      if ((j == pos) || (ci >= pyr.colsIn) || (ri >= pyr.rowsIn) || 
	  (pyr.getMaskIn() && !pyr.imgInMask.p(ci,ri))) continue;
      
      /* Add imgIndex if it does not already exist */
      for (i = index.begin(); i != index.end(); ++i) {
	if (*i == j) break;
	if (*i > j) {
	  index.insert(i,j);
	  break;
	}
      }
      if (i == index.end()) index.push_back(j);
    }
    }
    // Rerefernce imgIndex
    // Note: the first index of each list either points to itself, or a
    // list in which the first site has the same neighbourhood as itself.
    list<int> &imgIndex2 = pyrUp.imgInIndexl[*(pyrUp.imgInIndexl[*im].begin())];
    for (list<int>::const_iterator im2 = imgIndex2.begin(); im2 != imgIndex2.end(); ++im2) {
      int c = 2*(*im2%pyrUp.colsIn);
      int r = 2*(*im2/pyrUp.colsIn);
      for (int dy = 0; dy < 2; ++dy) {
      for (int dx = 0; dx < 2; ++dx) {
	int ci = c+dx;
	int ri = r+dy;
	j = ci+ri*pyr.colsIn;
	if ((j == pos) || (ci >= pyr.colsIn) || (ri >= pyr.rowsIn) || 
	    (pyr.getMaskIn() && !pyr.imgInMask.p(ci,ri))) continue;

	/* Add imgIndex if it does not already exist */
	for (i = index.begin(); i != index.end(); ++i) {
	  if (*i == j) break;
	  if (*i > j) {
	    index.insert(i,j);
	    break;
	  }
	}
	if (i == index.end()) index.push_back(j);
      }
      }
    }
  }
}

/**************************************************************************/

bool getImgPos(int &c, int &r, list<int> &index, int cols, int rows,
	       GreyImage &imgMask, bool mask) {
  static int i = 0;
  static list<int>::iterator id;
  
  if (index.empty()) {
    int size = cols*rows;
    if (i == size) {
      i = 0;
      return false;
    }
    c = i%cols;
    r = i/cols;
    ++i;
    while (mask && !imgMask.p(c,r)) {
      if (i == size) {
	i = 0;
	return false;
      }
      c = i%cols;
      r = i/cols;
      ++i;
    }
    return true;

  } else {
    if (i == 0) id = index.begin();
    else if (id == index.end()) {
      i = 0;
      return false;
    }
    c = (*id)%cols;
    r = (*id)/cols;
    ++i;
    ++id;
    return true;
  }
}

/**************************************************************************/

bool getImgPos2(int &c, int &r, int startIndex, list<int> &index,
		int cols, int rows, GreyImage &imgMask, bool mask) {
  static int i = 0;
  static list<int>::iterator id;
  
  if (index.empty()) {
    int size = cols*rows;
    if (i == 0) i = startIndex+1;
    if (i == size) {
      i = 0;
      return false;
    }
    c = i%cols;
    r = i/cols;
    ++i;
    while (mask && !imgMask.p(c,r)) {
      if (i == size) {
	i = 0;
	return false;
      }
      c = i%cols;
      r = i/cols;
      ++i;
    }
    return true;

  } else {
    if (i == 0) id = index.begin();
    else if (id == index.end()) {
      i = 0;
      return false;
    }
    c = (*id)%cols;
    r = (*id)/cols;
    ++i;
    ++id;
    return true;
  }
}

/**************************************************************************/
 
