#ifndef _BasicImage_h
#define _BasicImage_h

#include <iostream>
#include <string>
#include <Magick++.h>

using namespace std;

/**
 * Abstraction of an image.
 * Pixel access functionality is implemented in subclasses!
 *
 * Note: Pixel storage is implemented in subclasses.
 */
class BasicImage {

public:

  typedef unsigned char byte;

protected:
  int width, height, size;   // image dimensions
  int quantLevels;           // quantization levels

public:

  virtual ~BasicImage();

  virtual BasicImage *copy() const = 0;

  /**
   * Load image from disk.
   */
  virtual bool load(const string& fname) = 0;

  /**
   * Save image to disk.
   * The format is inferred from the suffix.
   */
  virtual bool save(const string& fname) const = 0;

  /**
   * Write image info (type, quantization, ...) to stream.
   */
  virtual void writeImageType(ostream& out) const;

  /**
   * Random shuffle of pixels
   */
  virtual void shuffle() = 0;

  /**
   * Clear image (to zero)
   */
  virtual void clear() = 0;

  /**
   * Get variables
   */
  inline virtual int getWidth() const;
  inline virtual int getHeight() const;
  inline virtual int getLevels() const;

  /**
   * Tests if a certain coordinate is within the image boundaries.
   */
  inline virtual bool contains(int x, int y) const;
  inline virtual bool containsF(float x, float y) const;

  /**
   * Wrap pixel points such that image is cyclic
   */
  inline virtual void cyclic(int& x, int& y) const;

protected:

  BasicImage();
  BasicImage(int w, int h, int lev);

  /**
   * Loads an image using Magick++, and assigns width/height
   */
  virtual bool loadMagick(const string& fname, Magick::Image& magImg);
  virtual bool saveMagick(const string& fname, Magick::Image& magImg) const;

  inline virtual void calcInterpCoords(float x, float y, float& xp, float& yp,
				       int& xl, int& yl,
				       int& xh, int& yh) const;

};


// ----- inline stuff -----

#include <cmath>

int BasicImage::getWidth() const { return width; }
int BasicImage::getHeight() const { return height; }
int BasicImage::getLevels() const { return quantLevels; }

bool BasicImage::contains(int x, int y) const {
  return (x >= 0 && x < width && y >= 0 && y < height);
}
bool BasicImage::containsF(float x, float y) const { // or a function template?
  return (x >= 0.0 && x <= width-1 && y >= 0.0 && y <= height-1);
}

void BasicImage::cyclic(int& x, int& y) const {
  if (x < 0) {
    x = x % width;
    if (x < 0) x += width;
  }
  else if (x >= width) x = x % width;
  if (y < 0) {
    y = y % height;
    if (y < 0) y += height;
  }
  else if (y >= height) y = y % height;
}

void BasicImage::calcInterpCoords(float x, float y, float& xp, float& yp,
				  int& xl, int& yl,
				  int& xh, int& yh) const {
  float xl_f = floor(x);
  float yl_f = floor(y);
  xp = x - xl_f;
  yp = y - yl_f;
  xl = (int)xl_f;
  yl = (int)yl_f;
  xh = (int)ceil(x);
  yh = (int)ceil(y);
}

#endif
