#include "GreyImage.h"

#include <Magick++.h>
#include <iostream>
#include <algorithm>
#include <string.h>

// using namespace std;

GreyImage::GreyImage(int w, int h, int lev) :
  BasicImage(w, h, lev), img(0), imgYX(0) {
  build();
}

GreyImage::GreyImage(const GreyImage& from) :
  BasicImage(from), img(0), imgYX(0) {
  build();
  if (size > 0) memcpy(img,from.img,size*sizeof(byte));
}

GreyImage::GreyImage(const string& fname, int lev) : img(0), imgYX(0) {
  load(fname,lev);
}

GreyImage::~GreyImage() {
  destroy();
}

BasicImage *GreyImage::copy() const {
  return new GreyImage(*this);
}

GreyImage const &GreyImage::operator= (GreyImage const &from) {
  if (this != &from) {
    destroy();
    copy(from);
  }
  return (*this);
}

void GreyImage::resize(int w, int h) {
  destroy();
  width = w;
  height = h;
  size = width*height;
  build();
}

void GreyImage::resize(int w, int h, int levels) {
  quantLevels = levels;
  resize(w,h);
}

bool GreyImage::load(const string& fname) {
  destroy();
  Magick::Image magImg;

  if (!loadMagick(fname,magImg)) return false;

  destroy();
  build();
  
  Magick::Blob bl;
  magImg.write(&bl, "GRAY", 8);
  memcpy(img,bl.data(),size*sizeof(byte));

  quantUniform();
  upsideDown();

  return true;
}

bool GreyImage::load(const string& fname, int lev) {
  quantLevels = lev;
  return load(fname);
}

bool GreyImage::save(const string& fname) const {
  GreyImage tmp(*this); // work with copy

  tmp.upsideDown();
  tmp.unquantUniform();

  Magick::Blob bl(tmp.img,size);
  Magick::Image magImg(bl, Magick::Geometry(width,height), 8, "GRAY");

  return saveMagick(fname,magImg);
}

void GreyImage::writeImageType(ostream& out) const {
  BasicImage::writeImageType(out);
  out << "image_mode " << "grey" << endl;
}

void GreyImage::shuffle() {
  //Factory::generator()->ShuffleVector(img, width*height);
  std::random_shuffle(img, img+size*sizeof(byte));
}

void GreyImage::clear() {
  memset(img, 0, size*sizeof(byte));
}

void GreyImage::copy(GreyImage const &from) {
  width = from.width;
  height = from.height;
  size = from.size;
  build();
  if (size > 0) memcpy(img,from.img,size*sizeof(byte));
}

void GreyImage::destroy() {
  delete [] img;
  delete [] imgYX;
  img = 0;
  imgYX = 0;
}

void GreyImage::build() {
  if (size > 0) {
    img = new byte[size];
    imgYX = new (byte *) [height];
    imgYX[0] = img;
    for (int y = 1; y < height; ++y) imgYX[y] = imgYX[y-1]+width;
  }
}

void GreyImage::quantUniform() {
  int f = 256/quantLevels;
  for (int i=0; i<size; i++) img[i] /= f;
}
void GreyImage::unquantUniform() {
  int f = 256/quantLevels;
  for (int i=0; i<size; i++) img[i] *= f;
}

void GreyImage::upsideDown() {
  byte *tmp = new byte[width];
  for (int y=0; y<height/2; y++) {
    int linelen = width*sizeof(byte);
    memcpy(tmp, img+y*width, linelen);
    memcpy(img+y*width, img+(height-1-y)*width, linelen);
    memcpy(img+(height-1-y)*width, tmp, linelen);
  }
  delete[] tmp;
}

GreyImage const &GreyImage::reduce(GreyImage const &from) {
  resize((from.width+1)/2,(from.height+1)/2);
  if (!width || !height) return (*this);
  float filter[5] = {0.05,0.25,0.4,0.25,0.05};
  float data;

  // REDUCE width
  float **imgTmpYX = new (float *) [from.height];
  imgTmpYX[0] = new float [width*from.height];
  for (int y = 1; y < from.height; ++y) imgTmpYX[y] = imgTmpYX[y-1]+width;
  for (int y = 0; y < from.height; ++y) {
    // x = 0
    // This results from using Burt and Adelson's Boundary conditions.
    int x = 0;
    imgTmpYX[y][0] = from.imgYX[y][0];

    // x = 1; x < width-1
    for (x = 1; x < width-1; ++x) {
      int xf = x*2;
      data = 0.0;
      for (int i = 0; i < 5; ++i)
	data += filter[i]*from.imgYX[y][xf+i-2];
      imgTmpYX[y][x] = data;
    }
    
    // x = width-1
    // This results from using Burt and Adelson's Boundary conditions.
    // (Well ok, not all the time)
    // but lets define the outer boundary edge so this is true)
    if (x < width) imgTmpYX[y][x] = from.imgYX[y][x*2];
  }

  // REDUCE height
  for (int x = 0; x < width; ++x) {
    // y = 0
    // This results from using Burt and Adelson's Boundary conditions.
    int y = 0;
    imgYX[0][x] = (byte)(imgTmpYX[0][x]+0.5);

    // y = 1; y < height-1
    for (y = 1; y < height-1; ++y) {
      int yf = y*2;
      data = 0.0;
      for (int i = 0; i < 5; ++i)
	data += filter[i]*imgTmpYX[yf+i-2][x];
      imgYX[y][x] = (byte)(data+0.5);
    }
    
    // y = height-1
    // This results from using Burt and Adelson's Boundary conditions.
    // (Well ok, not all the time,
    // but lets define the outer boundary edge so this is true)
    if (y < height) imgYX[y][x] = (byte)(imgTmpYX[y*2][x]+0.5);
  }

  delete [] imgTmpYX[0];
  delete [] imgTmpYX;
  return (*this);
}

GreyImage const &GreyImage::reduceX(GreyImage const &from) {
  resize((from.width+1)/2,height);
  float filter[5] = {0.05,0.25,0.4,0.25,0.05};
  float data;

  if (!width) return (*this);
  for (int y = 0; y < height; ++y) {
    // x = 0
    // This results from using Burt and Adelson's Boundary conditions.
    int x = 0;
    imgYX[y][0] = from.imgYX[y][0];

    // x = 1; x < width-1
    for (x = 1; x < width-1; ++x) {
      int xf = x*2;
      data = 0.0;
      for (int i = 0; i < 5; ++i)
	data += filter[i]*from.imgYX[y][xf+i-2];
      imgYX[y][x] = (byte)(data+0.5);
    }
    
    // x = width-1
    // This results from using Burt and Adelson's Boundary conditions.
    // (Well ok, not all the time)
    if (x < width) imgYX[y][x] = from.imgYX[y][x*2];
  }
  return (*this);
}

GreyImage const &GreyImage::reduceY(GreyImage const &from) {
  resize(width,(height+1)/2);
  float filter[5] = {0.05,0.25,0.4,0.25,0.05};
  float data;

  if (!height) return (*this);
  for (int x = 0; x < width; ++x) {
    // y = 0
    // This results from using Burt and Adelson's Boundary conditions.
    int y = 0;
    imgYX[0][x] = from.imgYX[0][x];

    // y = 1; y < height-1
    for (y = 1; y < height-1; ++y) {
      int yf = y*2;
      data = 0.0;
      for (int i = 0; i < 5; ++i)
	data += filter[i]*from.imgYX[yf+i-2][x];
      imgYX[y][x] = (byte)(data+0.5);
    }
    
    // y = height-1
    // This results from using Burt and Adelson's Boundary conditions.
    // (Well ok, not all the time,
    // but lets define the outer boundary edge so this is true)
    if (y < height) imgYX[y][x] = from.imgYX[y*2][x];
  }
  return (*this);
}

GreyImage const &GreyImage::sample(GreyImage const &from) {
  resize((from.width+1)/2,(from.height+1)/2);
  if (!width || !height) return (*this);
  for (int y = 0; y < height; ++y) {
    int yf = y*2;
    for (int x = 0; x < width; ++x) {
      imgYX[y][x] = from.imgYX[yf][x*2];
    }
  }
  return (*this);
}

BasicImage::byte GreyImage::pINearest(float x, float y) const {
  float xp,yp;
  int xl,yl,xh,yh;

  calcInterpCoords(x,y, xp,yp, xl,yl,xh,yh);

  // nearest neighbour
  if (xp < 0.5) {
    if (yp < 0.5) return p(xl, yl); else return p(xl, yh);
  } else {
    if (yp < 0.5) return p(xh, yl); else return p(xh, yh);
  }
}

BasicImage::byte GreyImage::pIBilinear(float x, float y) const {
  float xp,yp;
  int xl,yl,xh,yh;

  calcInterpCoords(x,y, xp,yp, xl,yl,xh,yh);

  int sw = p(xl, yl);
  int nw = p(xl, yh);
  int ne = p(xh, yh);
  int se = p(xh, yl);
  return (byte)(sw+(se-sw)*xp+(nw-sw+(sw-se-nw+ne)*xp)*yp + 0.5);
  //     (1.0-yp)*((1.0-xp)*sw+xp*se) + yp*((1.0-xp)*nw+xp*ne); //suboptimal
}
