#ifndef _GreyImage_h
#define _GreyImage_h

#include <Magick++.h>
#include <iostream>
#include <string>
#include "BasicImage.h"

using namespace std;

/**
 * An image of 1-byte grey values.
 *
 * Implementation note: Internally, stored "upside-down", i.e. first
 * pixels in memory correspond to line with maximal y coordinate.
 * Image memory is managed (creation, copy) by subclasses!
 *
 * This class relies on ImageMagick for loading/saving the images.
 */
class GreyImage : public BasicImage {

  friend class ColourImage; // ... needs to call protected methods

protected:
  byte *img;       // reference to pixel data, row by row
  byte **imgYX;    // reference to pixel data, row by row

public:

  /**
   * Constructs a greyscale image and optionally sets a uniform quantization.
   */
  GreyImage(int w = 0, int h = 0, int lev = 256);
  GreyImage(const string& fname, int lev = 256);
  GreyImage(const GreyImage& from);

  virtual ~GreyImage();

  virtual BasicImage *copy() const;

  virtual GreyImage const &operator= (GreyImage const &from);

  virtual void resize(int w, int h);
  virtual void resize(int w, int h, int levels);

  /**
   * Do Burt and Adelson's REDUCE
   * by applying a 5 point Gaussian filter
   */
  virtual GreyImage const &reduce(GreyImage const &from);
  virtual GreyImage const &sample(GreyImage const &from);

  /**
   * Greyscale intensity or index value at a certain position.
   */
  inline virtual BasicImage::byte p(int x, int y) const;
  inline virtual void setp(int x, int y, BasicImage::byte val);

  /**
   * Bilinear interpolation
   */
  inline virtual BasicImage::byte pI(float x, float y) const;

  /**
   * Two interpolation methods
   */
  virtual BasicImage::byte pINearest(float x, float y) const;
  virtual BasicImage::byte pIBilinear(float x, float y) const;

  /**
   * See BasicImage
   */
  virtual bool load(const string& fname);
  virtual bool load(const string& fname, int lev);
  virtual bool save(const string& fname) const;
  virtual void writeImageType(ostream& out) const;
  virtual void shuffle();
  virtual void clear();

protected:
  virtual void copy(GreyImage const &from);
  virtual void destroy();
  virtual void build();
  virtual void quantUniform();
  virtual void unquantUniform();
  virtual void upsideDown();
  virtual GreyImage const &reduceX(GreyImage const &from);
  virtual GreyImage const &reduceY(GreyImage const &from);
};


// ----- inline stuff -----

BasicImage::byte GreyImage::p(int x, int y) const {
  return imgYX[y][x];
}

void GreyImage::setp(int x, int y, BasicImage::byte val) {
  imgYX[y][x] = val;
}

BasicImage::byte GreyImage::pI(float x, float y) const {
  return pIBilinear(x,y);
}

#endif
